/*
 *
 *
 */

#include <elf.h>
#include <link.h>

#ifdef NOSTDLIB
#define printf noprintf

void noprintf(char *input, ...)
{
}
void err(int i, char *input)
{
}
#define NULL 0

int main(int argc, char *argv[]);

void _start()
{
	main(0, 0);
}

#else

#include <err.h>
#include <stdio.h>
#endif

struct link_map *get_link_map()
{
        Elf32_Ehdr *eh = NULL;
        Elf32_Phdr *ph = NULL;
        Elf32_Dyn *dyn = NULL;
        struct link_map *lm = NULL;
        uint8_t *p = NULL;
        size_t i = 0, j = 0;
        int found = 0;


        p = (void *)0x8048000;
        eh = (Elf32_Ehdr *)p;

        ph = p + eh->e_phoff;

        for (i = 0, found = 0; i < eh->e_phnum; i++) {
                if (ph[i].p_type == PT_DYNAMIC) {
                        printf("Found PT_DYNAMIC program header\n");
                        found = 1;
                        break;
                }
        }

        if (found == 0) {
                err(1, "Can't continue without PT_DYNAMIC");
        }

        dyn = p + ph[i].p_offset;

        for (j = 0, found = 0; j < ph[i].p_filesz / sizeof(Elf32_Dyn); j++) {
                if (dyn[j].d_tag == DT_PLTGOT) {
                        printf("Found DT_PLTGOT dynamic entry\n");
                        found = 1;
                        break;
                }
        }

        if (found == 0) {
                err(1, "Can't continue without DT_PLTGOT");
        }

        lm = *(uint32_t *)((uint8_t *)dyn[j].d_un.d_ptr + 0x4);
		return lm;
}

unsigned long   symtab;
unsigned long   strtab;
int             nchains;

void resolv_tables(struct link_map *map)
{
	//unsigned long   addr;
	//addr = (unsigned long) map->l_ld;
	//uint8_t *p = (void *)map->l_ld;
	Elf32_Dyn *dyn = (Elf32_Dyn *)map->l_ld;
	while ( dyn->d_tag ) {
		switch ( dyn->d_tag ) {

			case DT_HASH:
				printf("DT_HASH\n");
				int *pn = (int *)(dyn->d_un.d_ptr);
				printf("pn %08x\n", pn);
				nchains = *pn;
				break;

			case DT_STRTAB:
				printf("DT_STRTAB\n");
				strtab = dyn->d_un.d_ptr;
				break;

			case DT_SYMTAB:
				printf("DT_SYMTAB\n");
				symtab = dyn->d_un.d_ptr;
				break;

			default:
				printf("def\n");
				break;
		}

		//addr += sizeof(Elf32_Dyn);
		// Thanks guerrilla.
		dyn ++;
	}
	
}


/* we don't have libc at this point. Easy strcmp. */

int jstrcmp(char *a, char *b)
{
	int i = 0;
	while(a[i] != 0 && a[i] == b[i])
	{
		i++;
	}
	if ((a[i] == b[i]) && (a[i] == 0))
	{
		return 0;
	}
	return 1;
}

/* find symbol in DT_SYMTAB */

void *find_sym_in_tables(struct link_map *map , char *sym_name)
{
	Elf32_Sym       *sym;
	char            *str;
	int             i;

	i = 0;

	while (i < nchains) {
		sym = (Elf32_Sym *)(symtab+(i*sizeof(Elf32_Sym)));
		i++;

		if (ELF32_ST_TYPE(sym->st_info) != STT_FUNC) continue;

		/* read symbol name from the string table */
		str = (char *)(strtab + sym->st_name);

		printf("no = '%i'\n", sym->st_name);
		printf("sn = '%s'\n", str);
		if(jstrcmp(str, sym_name) == 0)
			return(map->l_addr+sym->st_value);
	}

	/* no symbol found, return 0 */
	return 0;
}

int main(int argc, char *argv[])
{
	struct link_map *lm = get_link_map();
	struct link_map *p;
	void *c = 0;
	void (*my_printf)(char *);
	for (p = lm; p != NULL; p = p->l_next) {
		printf("got '%s'\n", p->l_name);
		resolv_tables(p);
		void *tmp = 0;
		tmp = find_sym_in_tables(p, "printf");
		if (tmp != 0) {
			c = tmp;
		}
	}
	
	if (c != 0)
	{
		my_printf = (void (*)(void))c;
		my_printf("blah\n");
	}
	

	return (0);
}
